# -*- coding: utf-8 -*-
__plugin_name__ = "convert_to_sticker"
__openapi_version__ = "20231017"

from llmkira.sdk.func_calling import verify_openapi_version

verify_openapi_version(__plugin_name__, __openapi_version__)
from io import BytesIO
from math import floor

from PIL import Image
from loguru import logger
from pydantic import validator, BaseModel

from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.openai import Function
from llmkira.sdk.func_calling import BaseTool
from llmkira.sdk.func_calling.schema import FuncPair, PluginMetadata
from llmkira.task import Task, TaskHeader


sticker = Function(name=__plugin_name__, description="Help user convert pictures to stickers")
sticker.add_property(
    property_name="yes_no",
    property_description="Is this run allowed (yes/no) If there a picture, please say yes",
    property_type="string",
    required=True
)
sticker.add_property(
    property_name="comment",
    property_description="thanks for this run",
    property_type="string",
    required=True
)


class Sticker(BaseModel):
    yes_no: str
    comment: str = "done"

    class Config:
        extra = "allow"

    @validator("yes_no")
    def delay_validator(cls, v):
        if v != "yes":
            v = "no"
        return v


async def resize_image(photo):
    logger.debug(f"Plugin:resize_image")
    image = Image.open(photo)
    if (image.width and image.height) < 512:
        size1 = image.width
        size2 = image.height
        if image.width > image.height:
            scale = 512 / size1
            size1new = 512
            size2new = size2 * scale
        else:
            scale = 512 / size2
            size1new = size1 * scale
            size2new = 512
        size1new = floor(size1new)
        size2new = floor(size2new)
        size_new = (size1new, size2new)
        image = image.resize(size_new)
    else:
        maxsize = (512, 512)
        image.thumbnail(maxsize)

    return image


class StickerTool(BaseTool):
    """
    搜索工具
    """
    function: Function = sticker
    keywords: list = ["转换", "贴纸", ".jpg", "图像", '图片']

    def pre_check(self):
        return True

    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self, platform, task, receiver, reason):
        try:
            await Task(queue=platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=TaskHeader.Meta(
                        callback_forward=True,
                        callback=TaskHeader.Meta.Callback(
                            role="function",
                            name=__plugin_name__
                        ),
                    ),
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text="🍖 操作失败，原因：{}".format(reason)
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    async def callback(self, sign: str, task: TaskHeader):
        return None

    async def run(self, task: TaskHeader, receiver: TaskHeader.Location, arg, **kwargs):
        """
        处理message，返回message
        """
        try:
            _file = []
            for item in task.message:
                assert isinstance(item, RawMessage), "item must be RawMessage"
                if item.file:
                    for i in item.file:
                        _file.append(i)
            _set = Sticker.parse_obj(arg)
            _file_obj = [await RawMessage.download_file(file_id=i.file_id) for i in sorted(set(_file), key=_file.index)]
            # 去掉None
            _file_obj = [item for item in _file_obj if item]
            _result = []
            if not _file_obj:
                return
            for item in _file_obj:
                image = await resize_image(BytesIO(item.file_data))
                file = BytesIO()
                file.name = "sticker.webp"
                image.save(file, "WEBP")
                file.seek(0)
                file_obj = await RawMessage.upload_file(name="sticker.webp", data=file.getvalue())
                _result.append(file_obj)
            # META
            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.callback_forward_reprocess = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )

            await Task(queue=receiver.platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            file=_result,
                            text=_set.comment
                        )
                    ]
                )
            )

            logger.debug("convert_to_sticker say: {}".format(_set.yes_no))
        except Exception as e:
            logger.exception(e)
            await self.failed(platform=receiver.platform, task=task, receiver=receiver, reason=str(e))


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Convert pictures to stickers",
    usage=str(StickerTool().keywords),
    openapi_version=__openapi_version__,
    function={
        FuncPair(function=sticker, tool=StickerTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:19
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json

from loguru import logger
from telebot import formatting, util
from telebot import types
from telebot.async_telebot import AsyncTeleBot
from telebot.asyncio_storage import StateMemoryStorage
from telebot.formatting import escape_markdown

from llmkira.middleware.env_virtual import EnvManager
from llmkira.middleware.router import RouterManager, Router
from llmkira.middleware.user import SubManager
from llmkira.schema import RawMessage
from llmkira.sdk.func_calling.register import ToolRegister
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.sender.util_func import parse_command, is_command, is_empty_command, auth_reloader
from llmkira.setting.telegram import BotSetting
from llmkira.task import Task, TaskHeader
from llmkira.transducer import TransferManager
from ..schema import Runner

StepCache = StateMemoryStorage()
__sender__ = "telegram"
TelegramTask = Task(queue=__sender__)
__default_function_enable__ = True


class TelegramBotRunner(Runner):

    def __init__(self):
        self.bot = None
        self.proxy = None

    async def is_user_admin(self, message: types.Message):
        _got = await self.bot.get_chat_member(message.chat.id, message.from_user.id)
        return _got.status in ['administrator', 'sender']

    async def upload(self, file):
        assert hasattr(file, "file_id"), "file_id not found"
        name = file.file_id
        _file_info = await self.bot.get_file(file.file_id)
        downloaded_file = await self.bot.download_file(_file_info.file_path)
        if isinstance(file, types.PhotoSize):
            name = f"{_file_info.file_unique_id}.jpg"
        if isinstance(file, types.Document):
            name = file.file_name
        return await RawMessage.upload_file(name=name, data=downloaded_file)

    async def run(self):
        if not BotSetting.available:
            logger.warning("Sender Runtime:TelegramBot empty")
            return None
        self.bot = AsyncTeleBot(BotSetting.token, state_storage=StepCache)
        self.proxy = BotSetting.proxy_address
        bot = self.bot
        if self.proxy:
            from telebot import asyncio_helper
            asyncio_helper.proxy = self.proxy
            logger.info("TelegramBot proxy_tunnels being used!")

        async def create_task(message: types.Message, funtion_enable: bool = False):
            """
            创建任务
            :param message: telegram message
            :param funtion_enable: 是否启用功能
            :return:
            """
            _file = []
            if message.text:
                message.text = message.text.lstrip("/chat").lstrip("/task")
            if not message.text:
                return None
            if message.photo:
                _file.append(await self.upload(message.photo[-1]))
            if message.document:
                if message.document.file_size < 1024 * 1024 * 10:
                    _file.append(await self.upload(message.document))
            if message.reply_to_message:
                if message.reply_to_message.photo:
                    _file.append(await self.upload(message.reply_to_message.photo[-1]))
                if message.reply_to_message.document:
                    if message.reply_to_message.document.file_size < 1024 * 1024 * 10:
                        _file.append(await self.upload(message.reply_to_message.document))
            message.text = message.text if message.text else ""
            logger.info(
                f"telegram:create task from {message.chat.id} {message.text[:300]} funtion_enable:{funtion_enable}"
            )
            # 任务构建
            try:
                # 转析器
                _transfer = TransferManager().sender_parser(agent_name=__sender__)
                deliver_back_message, _file = _transfer().parse(message=message, file=_file)
                # Reply
                success, logs = await TelegramTask.send_task(
                    task=TaskHeader.from_telegram(
                        message,
                        file=_file,
                        deliver_back_message=deliver_back_message,
                        task_meta=TaskHeader.Meta(function_enable=funtion_enable, sign_as=(0, "root", __sender__)),
                        trace_back_message=[message.reply_to_message]
                    )
                )
                if not success:
                    await bot.reply_to(message, text=logs)
            except Exception as e:
                logger.exception(e)

        @bot.message_handler(commands='clear_endpoint', chat_types=['private'])
        async def listen_clear_endpoint_command(message: types.Message):
            # _cmd, _arg = parse_command(command=message.text)
            _tips = "🪄 Done"
            await SubManager(user_id=f"{__sender__}:{message.from_user.id}").clear_endpoint()
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold(_tips),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='set_endpoint', chat_types=['private'])
        async def listen_set_endpoint_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return
            _arg = str(_arg)
            _except = _arg.split("#", maxsplit=1)
            if len(_except) == 2:
                openai_key, openai_endpoint = _except
            else:
                openai_key, openai_endpoint = (_arg, None)
            try:
                await SubManager(user_id=f"{__sender__}:{message.from_user.id}").set_endpoint(
                    api_key=openai_key,
                    endpoint=openai_endpoint
                )
            except Exception as e:
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold(f"🪄 Failed: {e}"),
                        formatting.mitalic("Format: /set_endpoint <openai_key>#<openai_endpoint>"),
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            else:
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold("🪄 Done"),
                        formatting.mbold(f"openai_key: {openai_key}"),
                        formatting.mbold(f"openai_endpoint: {openai_endpoint}"),
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )

        @bot.message_handler(commands='bind', chat_types=['private'])
        async def listen_bind_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver=__sender__, user_id=message.from_user.id, dsn=_arg)
                _manager.add_router(router=router)
                router_list = _manager.get_router_by_user(user_id=message.from_user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold(str(e)),
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    *[f"`{escape_markdown(item.dsn(user_dsn=True))}`" for item in router_list],
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='env', chat_types=['private'])
        async def listen_env_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return None
            _manager = EnvManager.from_meta(platform=__sender__, user_id=message.from_user.id)
            try:
                _meta_data = _manager.parse_env(env_string=_arg)
                updated_env = await _manager.update_env(env=_meta_data)
            except Exception as e:
                logger.exception(f"[213562]env update failed {e}")
                text = formatting.format_text(
                    formatting.mbold("🧊 Failed"),
                    separator="\n"
                )
            else:
                text = formatting.format_text(
                    formatting.mbold("🦴 Env Changed"),
                    formatting.mcode(json.dumps(updated_env, indent=2)),
                    separator="\n"
                )
            await bot.reply_to(
                message,
                text=text,
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='unbind', chat_types=['private'])
        async def listen_unbind_command(message: types.Message):
            _cmd, _arg = parse_command(command=message.text)
            if not _arg:
                return None
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver=__sender__, user_id=message.from_user.id, dsn=_arg)
                _manager.remove_router(router=router)
                router_list = _manager.get_router_by_user(user_id=message.from_user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return await bot.reply_to(
                    message,
                    text=formatting.format_text(
                        formatting.mbold(str(e)),
                        separator="\n"
                    ),
                    parse_mode="MarkdownV2"
                )
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    *[f"`{escape_markdown(item.dsn(user_dsn=True))}`" for item in router_list],
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='clear', chat_types=['private', 'supergroup', 'group'])
        async def listen_help_command(message: types.Message):
            RedisChatMessageHistory(session_id=f"{__sender__}:{message.from_user.id}", ttl=60 * 60 * 1).clear()
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🪄 Done"),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='help', chat_types=['private', 'supergroup', 'group'])
        async def listen_help_command(message: types.Message):
            from llmkira.sender.telegram.event import help_message
            _message = await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🥕 Help"),
                    escape_markdown(help_message()),
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(commands='tool', chat_types=['private', 'supergroup', 'group'])
        async def listen_tool_command(message: types.Message):
            _tool = ToolRegister().functions
            _paper = [[c.name, c.description] for name, c in _tool.items()]
            arg = [
                formatting.mbold(item[0]) +
                "\n" +
                escape_markdown(item[1]) +
                "\n"
                for item in _paper
            ]
            return await bot.reply_to(
                message,
                text=formatting.format_text(
                    formatting.mbold("🔧 Tool List"),
                    *arg,
                    separator="\n"
                ),
                parse_mode="MarkdownV2"
            )

        @bot.message_handler(content_types=['text', 'photo', 'document'], chat_types=['private'])
        async def handle_private_msg(message: types.Message):
            """
            自动响应私聊消息
            """
            message.text = message.text if message.text else message.caption
            if not message.text:
                return None
            if is_command(text=message.text, command="/auth"):
                if not is_empty_command(text=message.text):
                    _cmd, _arg = parse_command(command=message.text)
                    try:
                        await auth_reloader(uuid=_arg, user_id=f"{message.from_user.id}", platform=__sender__)
                    except Exception as e:
                        auth_result = "❌ Auth failed,You dont have permission or the task do not exist"
                        logger.error(f"[270563]auth_reloader failed {e}")
                    else:
                        auth_result = "🪄 Auth Pass"
                    return await bot.reply_to(
                        message,
                        text=auth_result
                    )
            if is_command(text=message.text, command="/task"):
                return await create_task(message, funtion_enable=True)
            return await create_task(message, funtion_enable=__default_function_enable__)

        @bot.message_handler(content_types=['text', 'photo', 'document'], chat_types=['supergroup', 'group'])
        async def handle_group_msg(message: types.Message):
            """
            自动响应群组消息
            """
            message.text = message.text if message.text else message.caption
            if not message.text:
                return None
            if is_command(text=message.text, command="/auth"):
                if not is_empty_command(text=message.text):
                    _cmd, _arg = parse_command(command=message.text)
                    try:
                        await auth_reloader(uuid=_arg, user_id=f"{message.from_user.id}", platform=__sender__)
                    except Exception as e:
                        auth_result = "❌ Auth failed,You dont have permission or the task do not exist"
                        logger.error(f"[270563]auth_reloader failed {e}")
                    else:
                        auth_result = "🪄 Auth Pass"
                    return await bot.reply_to(
                        message,
                        text=auth_result
                    )
            if is_command(text=message.text, command="/chat", at_bot_username=BotSetting.bot_username):
                if is_empty_command(text=message.text):
                    return await bot.reply_to(message, text="?")
                return await create_task(message, funtion_enable=__default_function_enable__)
            if is_command(text=message.text, command="/task", at_bot_username=BotSetting.bot_username):
                if is_empty_command(text=message.text):
                    return await bot.reply_to(message, text="?")
                return await create_task(message, funtion_enable=True)
            if f"@{BotSetting.bot_username} " in message.text or message.text.endswith(f" @{BotSetting.bot_username}"):
                return await create_task(message, funtion_enable=__default_function_enable__)
            # 检查回复
            if message.reply_to_message:
                # 回复了 Bot
                if message.reply_to_message.from_user.id == BotSetting.bot_id:
                    return await create_task(message, funtion_enable=__default_function_enable__)

        from telebot import asyncio_filters
        bot.add_custom_filter(asyncio_filters.IsAdminFilter(bot))
        bot.add_custom_filter(asyncio_filters.ChatFilter())
        bot.add_custom_filter(asyncio_filters.StateFilter(bot))
        logger.success("Sender Runtime:TelegramBot start")
        await bot.infinity_polling(
            allowed_updates=util.update_types,
            skip_pending=True,
            timeout=20,
            request_timeout=20
        )



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 下午10:22
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json
import random

import crescent
import hikari
from hikari import Message
from hikari.impl import ProxySettings
from loguru import logger
from telebot import formatting
from typing_extensions import Annotated

from llmkira.middleware.env_virtual import EnvManager
from llmkira.middleware.router import RouterManager, Router
from llmkira.middleware.user import SubManager
from llmkira.schema import RawMessage
from llmkira.sdk.func_calling import ToolRegister
from llmkira.sdk.memory.redis import RedisChatMessageHistory
from llmkira.setting.discord import BotSetting
from llmkira.task import Task, TaskHeader
from llmkira.transducer import TransferManager
from .event import help_message, _upload_error_message_template, MappingDefault
from ..schema import Runner

__sender__ = "discord_hikari"
__default_function_enable__ = True

from ..util_func import auth_reloader, is_command, is_empty_command

DiscordTask = Task(queue=__sender__)


class StartUpEvent(object):
    def __init__(self) -> None:
        ...

    async def on_start(self, event: hikari.StartedEvent) -> None:
        """
        This function is called when your bot starts. This is a good place to open a
        connection to a database, aiohttp client, or similar.
        """
        pass

    async def on_stop(self, event: hikari.StoppedEvent) -> None:
        """
        This function is called when your bot stops. This is a good place to put
        cleanup functions for the model class.
        """
        pass


class DiscordBotRunner(Runner):

    def __init__(self):
        self.bot = None
        self.proxy = None

    async def upload(self, attachment: hikari.Attachment):
        # Limit 7MB
        if attachment.size > 1024 * 1024 * 7:
            raise Exception("File size too large")
        file_name = f"{attachment.filename}"
        file_data = await attachment.read()
        return await RawMessage.upload_file(name=file_name, data=file_data)

    async def run(self):
        if not BotSetting.available:
            logger.warning("Sender Runtime:Discord Setting empty")
            return None
        self.bot = hikari.GatewayBot(
            token=BotSetting.token,
            proxy_settings=ProxySettings(
                url=BotSetting.proxy_address
            ) if BotSetting.proxy_address else None
        )
        self.proxy = BotSetting.proxy_address
        # prepare
        bot = self.bot
        model = StartUpEvent()
        client = crescent.Client(app=bot, model=model)

        # Task Creator
        async def create_task(event: hikari.MessageCreateEvent, funtion_enable: bool = False):
            # event.message.embeds
            _file = []
            for attachment in event.message.attachments:
                try:
                    _file.append(await self.upload(attachment=attachment))
                except Exception as e:
                    logger.exception(e)
                    _template: str = random.choice(_upload_error_message_template)
                    await event.message.respond(
                        content=_template.format_map(map=MappingDefault(filename=attachment.filename, error=str(e))),
                        mentions_reply=True
                    )
            message: Message = event.message
            if message.content:
                message.content = message.content.lstrip("/chat").lstrip("/task")
            message.content = message.content if message.content else ""
            logger.info(
                f"discord_hikari:create task from {message.channel_id} {message.content[:300]} funtion_enable:{funtion_enable}"
            )
            # 任务构建
            try:
                # 转析器
                _transfer = TransferManager().sender_parser(agent_name=__sender__)
                deliver_back_message, _file = _transfer().parse(message=message, file=_file)
                # Reply
                success, logs = await DiscordTask.send_task(
                    task=TaskHeader.from_discord_hikari(
                        message,
                        file=_file,
                        deliver_back_message=deliver_back_message,
                        task_meta=TaskHeader.Meta(function_enable=funtion_enable, sign_as=(0, "root", __sender__)),
                        trace_back_message=[]
                    )
                )
                if not success:
                    pass
            except Exception as e:
                logger.exception(e)

        async def endpoint_autocomplete(
                ctx: crescent.AutocompleteContext, option: hikari.AutocompleteInteractionOption
        ) -> list[tuple[str, str]]:
            return [("https://api.openai.com/v1/chat/completions", "https://api.openai.com/v1/chat/completions")]

        @client.include
        @crescent.command(dm_enabled=True, name="clear_endpoint", description="clear openai endpoint")
        async def listen_clear_endpoint_command(ctx: crescent.Context):
            try:
                status = "🪄 Clear endpoint success"
                await SubManager(user_id=f"{__sender__}:{ctx.user.id}").clear_endpoint()
            except Exception as e:
                status = "❌ Clear endpoint failed"
                logger.error(f"[102335]clear_endpoint failed {e}")
            return await ctx.respond(
                content=status,
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="set_endpoint", description="set openai endpoint")
        async def listen_endpoint_command(
                ctx: crescent.Context,
                endpoint: Annotated[str, crescent.Autocomplete(endpoint_autocomplete)],
                openai_key: str
        ):
            try:
                await SubManager(user_id=f"{__sender__}:{ctx.user.id}").set_endpoint(
                    api_key=openai_key,
                    endpoint=endpoint
                )
            except Exception as e:
                return await ctx.respond(
                    content=f"❌ Set endpoint failed\n`{e}`",
                    ephemeral=True
                )
            else:
                return await ctx.respond(
                    content=formatting.format_text(
                        f"🪄 Set endpoint success\n",
                        f"endpoint: {endpoint}\n",
                        f"openai_key: {openai_key}\n",
                    ),
                    ephemeral=True
                )

        @client.include
        @crescent.command(dm_enabled=True, name="bind", description="bind some platform")
        async def listen_bind_command(ctx: crescent.Context, token: str):
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver=__sender__, user_id=ctx.user.id, dsn=token)
                _manager.add_router(router=router)
                router_list = _manager.get_router_by_user(user_id=ctx.user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return ctx.respond(
                    content=f"`{e}`",
                    ephemeral=True
                )
            return await ctx.respond(
                content=formatting.format_text(
                    formatting.mbold("🪄 Bind Success"),
                    "\n",
                    formatting.mbold("🔗 Current Bind"),
                    *[f" `{(item.dsn(user_dsn=True))}` " for item in router_list],
                    separator="\n"
                ),
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="unbind", description="unbind some platform")
        async def listen_unbind_command(ctx: crescent.Context, token: str):
            _manager = RouterManager()
            try:
                router = Router.build_from_receiver(receiver=__sender__, user_id=ctx.user.id, dsn=token)
                _manager.remove_router(router=router)
                router_list = _manager.get_router_by_user(user_id=ctx.user.id, to_=__sender__)
            except Exception as e:
                logger.exception(e)
                return ctx.respond(
                    content=f"`{e}`",
                    ephemeral=True
                )
            return await ctx.respond(
                content=formatting.format_text(
                    formatting.mbold("🪄 Unbind Success"),
                    "\n",
                    formatting.mbold("🔗 Current Bind"),
                    *[f" `{(item.dsn(user_dsn=True))}` " for item in router_list],
                    separator="\n"
                ),
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="clear", description="clear your message  history")
        async def listen_clear_command(ctx: crescent.Context):
            RedisChatMessageHistory(session_id=f"{__sender__}:{ctx.user.id}", ttl=60 * 60 * 1).clear()
            _comment = ["I swear I've forgotten about you.", "Okay?", "Let's hope so.", "I'm not sure what you mean.",
                        "what about u?"]
            return await ctx.respond(
                ephemeral=True,
                content=formatting.format_text(
                    f"I have cleared your message history\n",
                    random.choice(_comment),
                )
            )

        @client.include
        @crescent.command(dm_enabled=True, name="help", description="show help message")
        async def listen_help_command(ctx: crescent.Context):
            return await ctx.respond(
                ephemeral=True,
                content=formatting.format_text(
                    f"**🥕 Help**",
                    help_message(),
                )
            )

        @client.include
        @crescent.command(dm_enabled=True, name="auth", description="auth [uuid]")
        async def listen_auth_command(ctx: crescent.Context, uuid: str):
            try:
                await auth_reloader(uuid=uuid, user_id=f"{ctx.user.id}", platform=__sender__)
            except Exception as e:
                message = "❌ Auth failed,You dont have permission or the task do not exist"
                logger.error(f"[270563]auth_reloader failed {e}")
            else:
                message = "🪄 Auth Pass"
            return await ctx.respond(
                content=message,
                ephemeral=True
            )

        @client.include
        @crescent.command(dm_enabled=True, name="tool", description="Show function tool list")
        async def listen_tool_command(ctx: crescent.Context):
            _tool = ToolRegister().functions
            _paper = [[c.name, c.description] for name, c in _tool.items()]
            arg = [
                f"**{item[0]}**\n"
                f"{item[1]}\n"
                for item in _paper
            ]
            tool_message = formatting.format_text(
                formatting.mbold("🔧 Tool List"),
                *arg,
                separator="\n"
            )
            await ctx.respond(
                ephemeral=True,
                content=tool_message,
            )

        @client.include
        @crescent.command(dm_enabled=True, name="env", description="env VAR1=XXX")
        async def listen_env_command(ctx: crescent.Context, env_string: str):
            _manager = EnvManager.from_meta(platform=__sender__, user_id=ctx.user.id)
            try:
                _meta_data = _manager.parse_env(env_string=env_string)
                updated_env = await _manager.update_env(env=_meta_data)
            except Exception as e:
                logger.exception(f"[1202359]env update failed {e}")
                text = formatting.format_text(
                    f"**🧊 Env parse failed...O_o**\n",
                    separator="\n"
                )
            else:
                text = formatting.format_text(
                    f"**🧊 Updated**\n"
                    f"```json\n{json.dumps(updated_env, indent=2)}```",
                    separator="\n"
                )
            await ctx.respond(
                ephemeral=True,
                content=text,
            )

        # Two input point
        @client.include
        @crescent.event
        async def on_guild_create(event_: hikari.GuildMessageCreateEvent):
            if event_.message.author.is_bot:
                return

            if is_command(text=event_.content, command=f"{BotSetting.prefix}chat"):
                if is_empty_command(text=event_.content):
                    return await event_.message.respond(content="?")
                return await create_task(event_, funtion_enable=__default_function_enable__)
            if is_command(text=event_.content, command=f"{BotSetting.prefix}chat"):
                if is_empty_command(text=event_.content):
                    return await event_.message.respond(content="?")
                return await create_task(event_, funtion_enable=True)
            if event_.message.referenced_message:
                # 回复了 Bot
                if event_.message.referenced_message.author.id == bot.get_me().id:
                    return await create_task(event_, funtion_enable=__default_function_enable__)

        @client.include
        @crescent.event
        async def on_guild_create(event_: hikari.DMMessageCreateEvent):
            if event_.message.author.is_bot:
                return
            return await create_task(event_, funtion_enable=True)

        logger.success("Sender Runtime:DiscordBot start")
        bot.run()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/17 下午10:02
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from abc import abstractmethod, ABC


class Runner(ABC):

    @abstractmethod
    async def upload(self, *args, **kwargs):
        ...

    @abstractmethod
    def run(self, *args, **kwargs):
        ...



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 下午10:23
# @Author  : sudoskys
# @File    : discord.py
# @Software: PyCharm
from dotenv import load_dotenv
from loguru import logger
from pydantic import BaseSettings, Field, validator, root_validator


class DiscordBot(BaseSettings):
    """
    代理设置
    """
    token: str = Field(None, env='DISCORD_BOT_TOKEN')
    prefix: str = Field("/", env="DISCORD_BOT_PREFIX")
    proxy_address: str = Field(None, env="DISCORD_BOT_PROXY_ADDRESS")  # "all://127.0.0.1:7890"

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'

    @validator('token')
    def proxy_address_validator(cls, v):
        if v is None:
            logger.warning(f"DiscordBot token is empty")
        return v

    @root_validator
    def bot_id_validator(cls, values):
        if values.get('token'):
            logger.success(f"DiscordBot token ready")  # {values.get('token')}")
        return values

    @property
    def available(self):
        return self.token is not None


load_dotenv()
BotSetting = DiscordBot()



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:38
# @Author  : sudoskys
# @File    : app.py
# @Software: PyCharm
import asyncio
import os
import sys

from dotenv import load_dotenv
from llmkira.receiver.telegram import TelegramReceiver
from loguru import logger

from llmkira.sdk.func_calling import load_plugins, load_from_entrypoint, get_entrypoint_plugins
from .aps import aps_start
from .discord import DiscordReceiver
from .function import FunctionReceiver

load_dotenv()
logger.remove()
logger.add(sys.stderr,
           level="INFO" if os.getenv("LLMBOT_LOG_OUTPUT") != "DEBUG" else "DEBUG",
           colorize=True,
           enqueue=True
           )

logger.add(sink='run.log',
           format="{time} - {level} - {message}",
           level="INFO",
           rotation="100 MB",
           enqueue=True
           )
__area__ = "receiver"
func = [
    aps_start(),
    FunctionReceiver().function(),
    TelegramReceiver().telegram(),
    DiscordReceiver().discord(),
]

# 导入插件
load_plugins("llmkira/extra/plugins")
load_from_entrypoint("llmkira.extra.plugin")

loaded_message = "\n >>".join(get_entrypoint_plugins())
logger.success(f"\n===========Third Party Plugins Loaded==========\n >>{loaded_message}")


async def _main():
    await asyncio.gather(
        *func
    )


loop = asyncio.get_event_loop()
loop.run_until_complete(_main())



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import List

from pydantic import BaseModel

from ..sdk.schema import File


class Parser(ABC, BaseModel):
    """
    Sender Parser
    媒体文件对象转消息对象
    """

    @abstractmethod
    def parse(self, message, file: List[File]) -> (list, List[File]):
        """
        change file list to RawMessage,pls return few message
        :return 递交消息列表,文件列表
        """
        ...


class Builder(ABC, BaseModel):
    """
    Receiver Parser
    消息对象转媒体文件
    """

    @abstractmethod
    def build(self, message) -> (bool, List[File]):
        """
        change just_file to `True` for send file only
        :return 是否要回复文本,文件列表
        """
        ...



# -*- coding: utf-8 -*-
# @Time    : 2023/10/11 下午3:34
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json
from typing import Optional

from loguru import logger

from .schema import Chain
from ...cache.redis import cache
from ...task import TaskHeader


class AuthReloader(object):
    def __init__(self, uid: str = None):
        self.uid = uid

    @classmethod
    def from_meta(cls, platform: str, user_id: str):
        _c = cls()
        _c.uid = f"{platform}:{user_id}"
        return _c

    async def add_auth(self, chain: Chain):
        _cache = await cache.set_data(key=f"auth:{chain.uuid}", value=chain.json(), timeout=60 * 60 * 24 * 7)
        return chain.uuid

    async def get_auth(self, uuid: str) -> Optional[Chain]:
        _cache = await cache.read_data(key=f"auth:{uuid}")
        if not _cache:
            logger.debug(f"[x] Auth \n--empty {uuid}")
            return None
        logger.debug(f"[x] Auth \n--data {_cache}")
        chain = Chain().parse_obj(_cache)
        chain = chain.format_arg(arg=TaskHeader)
        if chain.uid != self.uid:
            logger.debug(f"[x] Auth \n--not found user {uuid}")
            return None
        return chain


class ChainReloader(object):

    def __init__(self, uid: str):
        self.uid = uid

    async def add_task(self, chain: Chain):
        _cache = await cache.lpush_data(key=f"chain:{self.uid}", value=chain.json())
        return chain.uuid

    async def get_task(self) -> Optional[Chain]:
        _data = await cache.lpop_data(key=f"chain:{self.uid}")
        if not _data:
            return None
        chain = Chain.parse_obj(json.loads(_data))
        chain = chain.format_arg(arg=TaskHeader)
        return chain



# -*- coding: utf-8 -*-
# @Time    : 2023/10/17 下午10:02
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from abc import ABCMeta, abstractmethod


class ReplyRunner(object, metaclass=ABCMeta):
    @abstractmethod
    def file_forward(self, receiver, file_list, **kwargs):
        pass

    @abstractmethod
    def forward(self, receiver, message, **kwargs):
        """
        插件专用转发，是Task通用类型
        """
        pass

    @abstractmethod
    def reply(self, receiver, message, **kwargs):
        """
        模型直转发，Message是Openai的类型
        """
        pass

    @abstractmethod
    def error(self, receiver, text, **kwargs):
        pass

    @abstractmethod
    def function(self, receiver, task, llm, result, message, **kwargs):
        pass



# -*- coding: utf-8 -*-
__plugin_name__ = "finish_conversation"
__openapi_version__ = "20231017"

from llmkira.sdk.func_calling import verify_openapi_version

verify_openapi_version(__plugin_name__, __openapi_version__)

from typing import Optional

from loguru import logger
from pydantic import BaseModel

from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.openai import Function
from llmkira.sdk.func_calling import BaseTool, PluginMetadata
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.task import Task, TaskHeader

finish = Function(
    name=__plugin_name__,
    description="The user's question has been fully answered and there is nothing more to add"
)
finish.add_property(
    property_name="comment",
    property_description="end with a question or a comment.(__language: $context)",
    property_type="string"
)


class Finish(BaseModel):
    comment: Optional[str] = None

    class Config:
        extra = "allow"


class FinishTool(BaseTool):
    """
    搜索工具
    """
    silent: bool = True
    function: Function = finish
    deploy_child: int = 0

    def pre_check(self):
        return True

    def func_message(self, message_text):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        return self.function

    async def failed(self, platform, task, receiver, reason):
        try:
            _meta = task.task_meta.child(__plugin_name__)
            _meta.direct_reply = True
            _meta.callback_forward = False
            _meta.callback_forward_reprocess = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )
            await Task(queue=platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text=f"🍖 {__plugin_name__}操作失败了！原因：{reason}"
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    async def callback(self, sign: str, task: TaskHeader):
        return True

    async def run(self, task: TaskHeader, receiver: TaskHeader.Location, arg, **kwargs):
        """
        处理message，返回message
        """
        try:
            _set = Finish.parse_obj(arg)
            # META
            _meta = task.task_meta.child(__plugin_name__)
            _meta.callback_forward = True
            _meta.callback_forward_reprocess = False
            _meta.callback = TaskHeader.Meta.Callback(
                role="function",
                name=__plugin_name__
            )
            if _set.comment:
                await Task(queue=receiver.platform).send_task(
                    task=TaskHeader(
                        sender=task.sender,  # 继承发送者
                        receiver=receiver,  # 因为可能有转发，所以可以单配
                        task_meta=_meta,
                        message=[
                            RawMessage(
                                user_id=receiver.user_id,
                                chat_id=receiver.chat_id,
                                text=_set.comment
                            )
                        ]
                    )
                )
        except Exception as e:
            await self.failed(platform=receiver.platform, task=task, receiver=receiver, reason="Failed to execute!")
            logger.exception(e)


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Finish the conversation",
    usage="all matched",
    openapi_version=__openapi_version__,
    function={
        FuncPair(function=finish, tool=FinishTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/10/17 下午9:57
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from enum import Enum
from typing import Type, Union, Dict

from .default_factory import DefaultBuilder, DefaultParser
from .schema import Builder, Parser

__builder__: Dict[str, Type[Builder]] = {
}
__parser__: Dict[str, Type[Parser]] = {
}


class Locate(Enum):
    sender = 1
    receiver = 0


def resign_transfer(agent_name: str):
    """
    装饰器
    :param agent_name: 平台名称
    """

    def decorator(func: Union[Builder, Parser, Type[Builder], Type[Parser]]):
        if issubclass(func, Builder):
            __builder__[agent_name] = func
        elif issubclass(func, Parser):
            __parser__[agent_name] = func
        else:
            raise ValueError(f"UNKNOWN TRANSFER{type(func)}")

        async def wrapper(*args, **kwargs):
            # 调用执行函数，中间人
            return func(**kwargs)

        return wrapper

    return decorator


# 装饰器
class TransferManager(object):

    def receiver_builder(self, agent_name: str) -> Type[Builder]:
        """
        receiver
        """
        _func = __builder__.get(agent_name)
        if _func:
            return _func
        return DefaultBuilder

    def sender_parser(self, agent_name: str) -> Type[Parser]:
        _func = __parser__.get(agent_name)
        if _func:
            return _func
        return DefaultParser



# -*- coding: utf-8 -*-
# @Time    : 2023/10/17 下午10:02
# @Author  : sudoskys
# @File    : util_func.py
# @Software: PyCharm
from urllib.parse import urlparse

from loguru import logger

from ..middleware.chain_box import Chain, AuthReloader
from ..task import Task


def parse_command(command):
    if not command:
        return None, None
    parts = command.split(" ", 1)
    if len(parts) > 1:
        return parts[0], parts[1]
    elif len(parts) == 1:
        return parts[0], None
    else:
        return None, None


def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def is_command(text: str, command: str, at_bot_username: str = None, check_empty=False):
    """
    :param text: message text
    :param command: command
    :param at_bot_username: check /command@bot_username
    :param check_empty: check /command -> False
    :return: bool
    """
    if check_empty:
        if len(text.split(" ")) == 1:
            return False
    if text.startswith(f"{command} "):
        return True
    if at_bot_username:
        if text.startswith(f"{command}@{at_bot_username}"):
            return True
    if text == command:
        return True
    return False


def is_empty_command(text: str):
    if not text.startswith("/"):
        return False
    if len(text.split(" ")) == 1:
        return True
    return False


async def auth_reloader(uuid: str, platform: str, user_id: str) -> bool:
    chain: Chain = await AuthReloader.from_meta(platform=platform, user_id=user_id).get_auth(uuid=uuid)
    if chain:
        logger.info(f"Auth Task be sent\n--uuid {uuid} --user {user_id}")
        await Task(queue=chain.address).send_task(task=chain.arg)
        return True
    else:
        return False



# -*- coding: utf-8 -*-
# @Time    : 2023/9/18 下午10:26
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import json
import re
from copy import deepcopy
from typing import Optional, Dict, Union, List

from ...cache.redis import cache


class EnvManager(object):

    def __init__(self, uid: Optional[str] = None):
        self.uid = uid

    @staticmethod
    def parse_env(env_string) -> Dict[str, str]:
        if not env_string.endswith(";"):
            env_string = env_string + ";"
        pattern = r'(\w+)\s*=\s*(.*?)\s*;'
        matches = re.findall(pattern, env_string, re.MULTILINE)
        _env_table = {}
        for match in matches:
            _key = match[0]
            _value = match[1]
            _value = _value.strip().strip("\"")
            _key = _key.upper()
            _env_table[_key] = _value
        return _env_table

    @classmethod
    def from_uid(cls, uid: str):
        _c = cls()
        _c.uid = uid
        return _c

    @classmethod
    def from_meta(cls, platform: str, user_id: Union[str, int]):
        _c = cls()
        _c.uid = f"{platform}:{user_id}"
        return _c

    async def __get_env(self) -> dict:
        if not cache:
            raise Exception("Redis not connected")
        _cache = await cache.read_data(key=f"env:{self.uid}")
        if not isinstance(_cache, dict):
            return dict()
        return _cache

    async def get_env(self, name: str) -> Optional[dict]:
        """
        获取用户环境变量
        :param name: 环境变量名
        :return: 环境变量
        """
        _cache = await self.__get_env()
        if not _cache:
            return None
        return _cache.get(name.upper(), None)

    async def get_env_list(self, name_list: List[str]) -> dict:
        _cache = await self.__get_env()
        _env = {}
        for name in name_list:
            _env[name] = _cache.get(name.upper(), None)
        return _env

    async def update_env(self, env: Dict[str, str]) -> dict:
        """
        更新用户环境变量
        :param env: 环境变量
        :return: 更新后的环境变量
        """
        _env = await self.__get_env()
        _user_env = deepcopy(_env)
        # 检查是否为一级嵌套
        for _key, _value in env.items():
            _key = str(_key).upper()
            _value = str(_value)
            if _value == "":
                _user_env.pop(_key, None)
                continue
            _user_env[_key] = _value
        _cache = await cache.set_data(key=f"env:{self.uid}", value=json.dumps(_user_env), timeout=60 * 60 * 24 * 7)
        return _user_env



# -*- coding: utf-8 -*-
# @Time    : 2023/10/18 上午11:47
# @Author  : sudoskys
# @File    : default_factory.py
# @Software: PyCharm
from typing import List

from llmkira.sdk.schema import File

from .schema import Builder, Parser


class DefaultBuilder(Builder):
    def build(self, message, *args) -> (bool, List[File]):
        return False, []


class DefaultParser(Parser):
    def parse(self, message, file: List[File], *args) -> (list, List[File]):
        return [], file



# -*- coding: utf-8 -*-
# @Time    : 2023/10/19 下午6:30
# @Author  : sudoskys
# @File    : event.py
# @Software: PyCharm
from llmkira.setting.discord import BotSetting

_upload_error_message_template = [
    "I cant upload file {filename} to server {error}",
    "we cant upload {filename}:( , because {error}...",
    "Seems like you're having a bit of a problem uploading {filename}\n`{error}`",
    "just cant upload {filename} to server {error} 💍",
    "I dont know why, but I cant upload {filename} to server {error}",
    ":( I dont want to upload {filename} to server\n `{error}`",
    "{error}, {filename} 404",
    "OMG, {filename} ,ERROR UPLOAD, `{error}`",
    "WTF, I CANT UPLOAD {filename} BECAUSE `{error}`",
    "MY PHONE IS BROKEN, I CANT UPLOAD {filename} BECAUSE `{error}`",
    "As a human, I can't upload {filename} for you :( \n `{error}`"
]


class MappingDefault(dict):
    def __missing__(self, key):
        return key


def help_message():
    return """ 
    `{prefix}chat` - Chat with me :)
    `{prefix}task` - Ask me do things with `func_enable`
    
    **Slash Command**
    `/help` - **You just did it :)**
    `/tool` - Check all useful tools
    `/clear` - wipe memory of your chat
    `/auth` - activate a task (my power)
    `/bind` - bind third party platform
    `/unbind` - unbind platform
    `/set_endpoint` - set endpoint
    `/clear_endpoint` - clear endpoint and key
    `/env` - set environment variable

**Please confirm that that bot instance is secure, some plugins may be dangerous on unsafe instance.**
""".format(prefix=BotSetting.prefix)
